package gov.va.med.esr.service.impl;


import gov.va.med.esr.common.model.ee.IRS1095B;
import gov.va.med.esr.common.model.ee.MECPeriod;
import gov.va.med.esr.common.model.lookup.Country;
import gov.va.med.esr.common.model.lookup.TransmissionStatus;
import gov.va.med.esr.common.model.messaging.IrsTransmissionLogData;
import gov.va.med.esr.common.model.messaging.IrsTransmissionLogDetailEntry;
import gov.va.med.esr.common.model.messaging.IrsTransmissionLogEntry;
import gov.va.med.esr.common.model.party.SimpleAddress;
import gov.va.med.esr.service.IRSWebServiceDelegate;
import gov.va.med.esr.service.LookupService;
import gov.va.med.fw.persistent.DAOOperations;
import gov.va.med.fw.service.AbstractComponent;
import gov.va.med.fw.service.ServiceException;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringReader;
import java.io.StringWriter;
import java.math.BigInteger;
import java.net.URL;
import java.rmi.RemoteException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.TimeZone;
import java.util.UUID;

import javax.activation.DataHandler;
import javax.activation.FileDataSource;
import javax.xml.bind.JAXB;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeConstants;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.XMLGregorianCalendar;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.Holder;
import javax.xml.ws.soap.SOAPBinding;

import org.apache.axis.AxisFault;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.cxf.endpoint.Client;
import org.apache.cxf.frontend.ClientProxy;
import org.apache.cxf.transport.common.gzip.GZIPFeature;
import org.apache.cxf.transport.common.gzip.GZIPOutInterceptor;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;

import us.gov.treasury.irs.common.BinaryFormatCodeType;
import us.gov.treasury.irs.common.ErrorMessageDetail;
import us.gov.treasury.irs.common.ErrorMessageDetailType;
import us.gov.treasury.irs.common.TINRequestTypeCodeType;
import us.gov.treasury.irs.ext.aca.air._7.ACABulkBusinessHeaderRequestType;
import us.gov.treasury.irs.ext.aca.air._7.ACABulkReqTrnsmtStsReqGrpDtl;
import us.gov.treasury.irs.ext.aca.air._7.ACABulkReqTrnsmtStsRespGrpDtl;
import us.gov.treasury.irs.ext.aca.air._7.ACATransmitterManifestReqDtl;
import us.gov.treasury.irs.ext.aca.air._7.BusinessAddressGrpType;
import us.gov.treasury.irs.ext.aca.air._7.BusinessNameType;
import us.gov.treasury.irs.ext.aca.air._7.CompanyInformationGrp;
import us.gov.treasury.irs.ext.aca.air._7.CorrectedRecordInfoGrpType;
import us.gov.treasury.irs.ext.aca.air._7.EmployerCoveredIndividualType;
import us.gov.treasury.irs.ext.aca.air._7.ForeignAddressGrp;
import us.gov.treasury.irs.ext.aca.air._7.Form1094BUpstreamDetailType;
import us.gov.treasury.irs.ext.aca.air._7.Form1095BUpstreamDetailType;
import us.gov.treasury.irs.ext.aca.air._7.IssuerInfoGrpType;
import us.gov.treasury.irs.ext.aca.air._7.MonthIndGrpType;
import us.gov.treasury.irs.ext.aca.air._7.OtherCompletePersonNameType;
import us.gov.treasury.irs.ext.aca.air._7.ResponsibleIndividualGrpType;
import us.gov.treasury.irs.ext.aca.air._7.TransmissionTypeCdType;
import us.gov.treasury.irs.ext.aca.air._7.TransmitterErrorDetailGrpType;
import us.gov.treasury.irs.ext.aca.air._7.USAddressGrp;
import us.gov.treasury.irs.ext.aca.air._7.VendorInformationGrp;
import us.gov.treasury.irs.ext.aca.air._7.StateType;
import us.gov.treasury.irs.ext.aca.air._7.TransmissionStatusCodeType;
import us.gov.treasury.irs.msg.form1094_1095bctransmittermessage.FormBCTrnsmtSubmissionDtlType;
import us.gov.treasury.irs.msg.form1094_1095btransmitterupstreammessage.Form109495BTrnsmtUpstreamType;
import us.gov.treasury.irs.msg.irsacabulkrequesttransmitter.ACABulkRequestTransmitter;
import us.gov.treasury.irs.msg.irsacabulkrequesttransmitter.ACABulkRequestTransmitterResponse;
import us.gov.treasury.irs.msg.irstransmitterstatusrequest.ACABulkRequestTransmitterStatusDetailRequest;
import us.gov.treasury.irs.msg.irstransmitterstatusrequest.ACABulkRequestTransmitterStatusDetailResponse;
import us.gov.treasury.irs.srv.acabulkrequesttransmitterstatus.ACATransmitterStatusReqPortType;
import us.gov.treasury.irs.srv.acabulkrequesttransmitterstatus.GetACATransmitterStatusReqService;
import us.gov.treasury.irs.srv.gettransmitterbulkrequest.*;


public class IRSWebServiceDelegateImpl  extends AbstractComponent implements IRSWebServiceDelegate {


	private static final long serialVersionUID = 8534310803323475694L;
	private static final String EIN = "741612229";
	private static final String LEGAL_NAME = "DEPARTMENT OF VETERANS AFFAIRS";
	private static final String BUSINESS_TYPE = "FGAN";
	private static final String BUSINESS_STREET_ADDRESS = "1800 G ST NW RM 402B";
	private static final String BUSINESS_CITY = "WASHINGTON";
	private static final String BUSINESS_STATE = "DC";
	private static final String BUSINESS_ZIP = "20460";
	private static final String BUSINESS_PHONE = "(202)271-5400";
	private static final String FORM_TYPE_CODE = "1094/1095B";
	private static final String APP_ID = "SYS12";
	private static final String TYPE_CODE = "T";


	private static final String SUBMIT_ORIGINAL = "O";
	private static final String SUBMIT_CORRECTION = "C";
	private static final String SUBMIT_REPLACEMENT = "R";
	private static final BinaryFormatCodeType TYPE_XML = BinaryFormatCodeType.APPLICATION_XML;

	//always only one 1094 container
	private static final BigInteger PAYER_COUNT = BigInteger.valueOf(1);

	private String irsServiceBulkEndPoint = null;
	private String irsServiceStatusEndPoint = null;

	private LookupService lookupService = null;

	private String irsProcessingCode = null;

	private String tempFileLocation = null;

	//path and name to the cert file to be used for digitally signing the soap header
	private String irsSignatureFile = null;

	private String transmissionControlCode = null;
	private String softwareId = null;
	private String clientSystemId = null;
	private String irsContactFname = null;
	private String irsContactLname = null;
	private String irsContactPhone = null;

	private DAOOperations genericDAO;



	/**
	 * Main entry for original submission bulk entries to the IRS
	 * You should not invoke this service directly from outside context of the IRSTransmissionService
     * @param transmission log entry
     */
	 public String transmitOriginal(IrsTransmissionLogEntry transmission) throws ServiceException {

		 return generateTransmitterBulkSubmission(transmission, null);

	 }

	 /*
	  * Standard business header used by both the bulk submits and the Ack service.
	  */
	 private ACABulkBusinessHeaderRequestType getBusinessHeader() throws ServiceException {

		 us.gov.treasury.irs.ext.aca.air._7.ObjectFactory f = new us.gov.treasury.irs.ext.aca.air._7.ObjectFactory();
		 ACABulkBusinessHeaderRequestType bHeader = f.createACABulkBusinessHeaderRequestType();

		 SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
		 sdf.setTimeZone(TimeZone.getTimeZone("GMT"));
		 String date = sdf.format(new Date());
		 XMLGregorianCalendar xmlCal = null;

		try {
			xmlCal = DatatypeFactory.newInstance().newXMLGregorianCalendar(date);
		} catch (DatatypeConfigurationException e) {
			logger.error("Data Type Error in IRS Transmit:", e);
		}

		 UUID id = UUID.randomUUID();
		 String utid = id + ":" + APP_ID + ":" + this.getTransmissionControlCode() + "::" + TYPE_CODE;

		 bHeader.setTimestamp(xmlCal);
		 bHeader.setUniqueTransmissionId(utid);

		 return bHeader;
	 }


	 /*
	  * primary worker for constructing the IRS bulk submit payload from the
	  * transmission log entry objects, that have been pre-committed
	  *
	  */
	 private String generateTransmitterBulkSubmission(IrsTransmissionLogEntry transmission, String originalReceipt) throws ServiceException{

		 try {

			 //data file attachment, root container for all 1094 and 1095B form entries
			 ACABulkRequestTransmitter filebody = new ACABulkRequestTransmitter();


			 ACATransmitterManifestReqDtl headerManifest = new ACATransmitterManifestReqDtl();

			 setCommonHeader(transmission, headerManifest);

			 //send the original receipt if correction or replacing a previous transmisison
			 if (!transmission.getCorrectionInd().equalsIgnoreCase(SUBMIT_ORIGINAL) && originalReceipt != null) {
				 headerManifest.setOriginalReceiptId(StringUtils.split(originalReceipt, "|")[0]);
			 }


			 //get request data xml and set as mtom byte array
			 //1094<form type>_Request_<TCC>_<Date>T<TimeStamp>Z.xml
			 //1094B_Request_BY01G_20140101T010102000Z.xml
			 SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd'T'HHmmss'Z'");
			 sdf.setTimeZone(TimeZone.getTimeZone("GMT"));
			 String date = sdf.format(new Date());

			 String fileName = "1094B_Request_" + this.getTransmissionControlCode() + "_" + date + ".xml";

			 //don't like creating an actual tmp xml file, but content type meta data must contain a fileName
			 //and matched to docSystemFileName when validated by IRS
			 //get rid of this if there is a way to set a filename
			 //on the binary handler without it ever existing first on a file system
			 File tmp = new File(this.getTempFileLocation() + fileName);

			 if (transmission.getIrsTransmissionLogData() != null && transmission.getIrsTransmissionLogData().size() > 0 &&
					 transmission.getIrsTransmissionLogData().iterator().next().getRequestData() != null) {

				 //the form data is pre-constructed and comitted prior to this method
				 //write out the xml to file and get the file as binary attachment,
				 //file gets deleted right after this service call completes.
				 FileUtils.writeStringToFile(tmp, transmission.getIrsTransmissionLogData().iterator().next().getRequestData(), "UTF-8");

				 DataHandler handler = new DataHandler(new FileDataSource(tmp));

				 //filebody.setBulkExchangeFile(IOUtils.toByteArray(handler.getInputStream()));
				 filebody.setBulkExchangeFile(handler);

				 //compute checksum and size
				 headerManifest.setChecksumAugmentationNum(createChecksum(tmp.getAbsolutePath()));
				 headerManifest.setAttachmentByteSizeNum(BigInteger.valueOf(tmp.length()));

			 } else {
				 //we have no data somehow, no point to continue
				 throw new ServiceException("Form Data file does not exist in IRS Bulk Submission Request");
			 }

			 headerManifest.setDocumentSystemFileNm(fileName);

			 ACABulkBusinessHeaderRequestType busHeaderDetail = new ACABulkBusinessHeaderRequestType();
			 busHeaderDetail = getBusinessHeader();
			 Holder<ACABulkBusinessHeaderRequestType> holder = new Holder<ACABulkBusinessHeaderRequestType>(busHeaderDetail);


			 //TODO add WS-security header
			 Security wsse = new Security();

			 String responseStatus = "";
			 String errorMsg = "";
			 String status = "";
			 String receiptId = "";

			 URL wsdlURL = new URL(this.getIrsServiceBulkEndPoint() + "?wsdl");

		 	 BulkRequestTransmitterService irsTransmitSvc = new BulkRequestTransmitterService(wsdlURL);

		 	 GZIPFeature gz = new GZIPFeature();
		 	 gz.setThreshold(0);

		 	 BulkRequestTransmitterPortType port = irsTransmitSvc.getBulkRequestTransmitterPort(gz);

		 	 BindingProvider bp = (BindingProvider) port;
	         SOAPBinding binding = (SOAPBinding) bp.getBinding();
	         binding.setMTOMEnabled(true);


			ACABulkRequestTransmitterResponse response =
					new ACABulkRequestTransmitterResponse();
			try {
				response = port.bulkRequestTransmitter(filebody, null, wsse, holder, headerManifest);
				//response = port.bulkRequestTransmitter(filebody, null, null, holder, headerManifest);
				//response = port.bulkRequestTransmitter(filebody);
				//response = mockBulkResponse();

				//remove the tmp xml file after send
				tmp.delete();

				if (response == null)
					throw new ServiceException("IRS Transmit Request Failure: No Response Recieved");

			} catch (Exception ex) {
				StringWriter sw = new StringWriter();
				ex.printStackTrace(new PrintWriter(sw));
				String exceptionDetails = sw.toString();
				logger.error("Exception during IRS bulk request service call:" + exceptionDetails);
				throw new ServiceException("IRS Transmit Request Failure:" + ex.getMessage());
			}

			receiptId = response.getReceiptId();
			if (response.getErrorMessageDetail() != null) {
				ErrorMessageDetail error = response.getErrorMessageDetail();
				if (error != null) {
					errorMsg += error.getErrorMessageCd() + ":";
					errorMsg += error.getErrorMessageTxt();
				}
			}
			TransmissionStatusCodeType c = response.getTransmissionStatusCd();
			status = c.value();

			responseStatus = receiptId + "^" + status + "^" + errorMsg;

			return responseStatus;

		 } catch (AxisFault e) {
			logger.error("Axis Error in IRS Transmit:" + e.getMessage());
		} catch (NoSuchAlgorithmException e) {
			logger.error("NSA Error in IRS Transmit:" + e.getMessage());
		} catch (RemoteException e) {
			logger.error("Remote Error in IRS Transmit:" + e.getMessage());
		} catch (IOException e) {
			logger.error("IO Error writing temp file in IRS bulk transmit:" + e.getMessage());
		}

		 return null;
	 }


	 /*
	  * Mock response for testing or internal validation only
	  * TODO move these mocks to external class and load through spring with configurable params
	  */
	 private ACABulkRequestTransmitterResponse mockBulkResponse() {

		 ACABulkRequestTransmitterResponse response =
				 new ACABulkRequestTransmitterResponse();

		 final Random rnd = new Random();
		 String res = "1095B-15-" + String.valueOf(rnd.nextInt());
		 response.setReceiptId(res);

		 response.setTransmissionStatusCd(TransmissionStatusCodeType.PROCESSING);

		 ErrorMessageDetail d = new ErrorMessageDetail();
		 d.setErrorMessageCd("AIR101589");
		 d.setErrorMessageTxt("Invalid TCC");
		 response.setErrorMessageDetail(d);

		 return response;
	 }

	 /*
	  * Utility for setting MD5 on the form data file
	  */
	 private String createChecksum(String filename) throws NoSuchAlgorithmException, IOException {

	   FileInputStream fis =  new FileInputStream(filename);

	   byte[] buffer = new byte[1024];
	   MessageDigest complete = MessageDigest.getInstance("MD5");
	   int numRead;

	   do {
	    numRead = fis.read(buffer);
	    if (numRead > 0) {
	      complete.update(buffer, 0, numRead);
	      }
	    } while (numRead != -1);
	   fis.close();

	   byte[] b = complete.digest();

	   String result = "";

       for (int i=0; i < b.length; i++) {
           result += Integer.toString( ( b[i] & 0xff ) + 0x100, 16).substring( 1 );
       }
       return result;

	 }


	/*
	 * generic marshaling utility for any of the jax object factories
	 */
	 private String buildXmlDocument(JAXBElement document) {

		  Class<?> clazz = document.getValue().getClass();
		  try {
		    JAXBContext context =
		        JAXBContext.newInstance(clazz.getPackage().getName());
		    Marshaller m = context.createMarshaller();
		    m.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
		    StringWriter sw = new StringWriter();
		    m.marshal(document, sw);
		    String xml = sw.toString();

		    return xml;

		  } catch (JAXBException e) {
		    logger.error("Error building xml doc " + e);

		  }
		  return null;
	}

 	/**
	 * Main entry for replacing failed bulk batch submissions
	 * You should not invoke this service directly from outside context of the IRSTransmissionService
     * @param transmission log entry
     * @param original receipt of the batch that is being replaced
     */
	 public String retransmitBatchSubmission(IrsTransmissionLogEntry transmission, String originalReceipt) throws ServiceException {
		 //entire batch was expired or rejected, retransmit with original 1095B data and original receipt
		 return generateTransmitterBulkSubmission(transmission, originalReceipt);

	 }

 	/**
	 * Main entry for replacing failed individual person submissions
	 * You should not invoke this service directly from outside context of the IRSTransmissionService
     * @param transmission log entry
     * @param original unique record id of the person entry that is being replaced
     */
	 public String retransmitPersonSubmission(IrsTransmissionLogEntry transmission, String originalReceipt) throws ServiceException {
		 return generateTransmitterBulkSubmission(transmission, originalReceipt);

	 }


 	/**
	 * Main entry for requesting acknowledgment of a previous submitted entry in processing status
	 * You should not invoke this service directly from outside context of the IRSTransmissionService
     * @param transmission log entry
     */
	public void requestProcessingStatus(IrsTransmissionLogEntry transmission) throws ServiceException {


		if (transmission == null || transmission.getReceiptId() == null || transmission.getReceiptId().isEmpty()) {
			return;
		}

		String receipt = transmission.getReceiptId();

		try {

			//only the business header needed for acks

			ACABulkBusinessHeaderRequestType busHeaderDetail = getBusinessHeader();

			ACABulkRequestTransmitterStatusDetailRequest requestBody = new ACABulkRequestTransmitterStatusDetailRequest();

			ACABulkReqTrnsmtStsReqGrpDtl reqDetail = new ACABulkReqTrnsmtStsReqGrpDtl();

			reqDetail.setReceiptId(receipt);

			requestBody.setACABulkReqTrnsmtStsReqGrpDtl(reqDetail);

			ACABulkRequestTransmitterStatusDetailResponse response =
					new ACABulkRequestTransmitterStatusDetailResponse();

			URL wsdlURL = new URL(this.getIrsServiceStatusEndPoint() + "?wsdl");
			GetACATransmitterStatusReqService irsAckService = new GetACATransmitterStatusReqService(wsdlURL);
			ACATransmitterStatusReqPortType port = irsAckService.getACATransmitterStatusReqPort();
		    Holder<ACABulkBusinessHeaderRequestType> holder = new Holder<ACABulkBusinessHeaderRequestType>(busHeaderDetail);

		    //	TODO Add security header
		    Security wsse = new Security();


			try {
				// if ack request fails or no-response, just log it
				//eat the exception and consider as still processing
				//it will keep trying again each ack interval until it reaches expired status
				//response = mockResponse(receipt);
				response = port.getACATransmitterStatusReqOperation(requestBody, null, wsse, holder);

				if (response == null || response.getACABulkRequestTransmitterResponse() == null ||
						response.getACABulkRequestTransmitterResponse().getTransmissionStatusCd() == null) {
					logger.error("Exception during IRS request acknowlegement service call: No Response Returned");
					return;
				}
			} catch (Exception ex) {

				 StringWriter sw = new StringWriter();
				 ex.printStackTrace(new PrintWriter(sw));
				 String exceptionDetails = sw.toString();
				 logger.error("Exception during IRS request acknowlegement service call:" + exceptionDetails);
				return;
			}

			TransmissionStatusCodeType respType =
					response.getACABulkRequestTransmitterResponse().getTransmissionStatusCd();

			//transmission is already set to processing status by default
			//if not complete yet do nothing, it will check again in the next interval
			if (respType == TransmissionStatusCodeType.PROCESSING) {
				return;
			}

			//check each possible status return type
			if (respType == TransmissionStatusCodeType.ACCEPTED ||
					respType == TransmissionStatusCodeType.PARTIALLY_ACCEPTED) {
				transmission.setTransmissionStatus((TransmissionStatus) this.lookupService.getByCode(
						TransmissionStatus.class, TransmissionStatus.ACCEPTED.getCode()));

				transmission.setErrorDescription(null);
				transmission.setAckDate(new Date());
				transmission.setAckId(response.getACABulkRequestTransmitterResponse().getReceiptId());
			} else if (respType == TransmissionStatusCodeType.NOT_FOUND) {
				transmission.setTransmissionStatus((TransmissionStatus) this.lookupService.getByCode(
						TransmissionStatus.class, TransmissionStatus.NOT_FOUND.getCode()));

				transmission.setErrorDescription("Reqiested Receipt ID Not Found");
				transmission.setAckDate(new Date());
				transmission.setAckId(null);
			} else if (respType == TransmissionStatusCodeType.REJECTED) {
				transmission.setTransmissionStatus((TransmissionStatus) this.lookupService.getByCode(
						TransmissionStatus.class, TransmissionStatus.REJECTED.getCode()));

				//error may be here or in the data file attachment; or both
				if (response.getACABulkRequestTransmitterResponse().getErrorMessageDetail() != null) {
					transmission.setErrorDescription(response.getACABulkRequestTransmitterResponse().getErrorMessageDetail().getErrorMessageCd()
							+ ":" + response.getACABulkRequestTransmitterResponse().getErrorMessageDetail().getErrorMessageTxt());
				}

				transmission.setAckDate(new Date());
				transmission.setAckId(response.getACABulkRequestTransmitterResponse().getReceiptId());

				if (response.getACABulkReqTrnsmtStsRespGrpDtl() != null) {

					DataHandler dh = response.getACABulkReqTrnsmtStsRespGrpDtl().getBulkExchangeFile();
					//may or may not have an error data file attached, if so it has to be 1094 entry error
					if (dh != null) {
						List<TransmitterErrorDetailGrpType> errorList = getErrorsFromBulkFile(dh, transmission);
						if (errorList != null && errorList.size() > 0) {
							ErrorMessageDetailType firstErr = errorList.iterator().next().getErrorMessageDetail().get(0);
			    			transmission.setErrorDescription(transmission.getErrorDescription() + ": " +
			    					"Detail :" + firstErr.getErrorMessageCd() + ":" + firstErr.getErrorMessageTxt());

			    			if (transmission.getErrorDescription().length() > 900) {
			    				transmission.setErrorDescription(transmission.getErrorDescription().substring(0, 900) + " :TRUNCATED");
			    			}
						}
					}
				}

			} else if (respType == TransmissionStatusCodeType.ACCEPTED_WITH_ERRORS) {
				transmission.setTransmissionStatus((TransmissionStatus) this.lookupService.getByCode(
						TransmissionStatus.class, TransmissionStatus.ACCEPTED_WITH_ERRORS.getCode()));

				transmission.setErrorDescription(null);
				transmission.setAckDate(new Date());
				transmission.setAckId(response.getACABulkRequestTransmitterResponse().getReceiptId());

				DataHandler dh = response.getACABulkReqTrnsmtStsRespGrpDtl().getBulkExchangeFile();

				List<TransmitterErrorDetailGrpType> errorList = getErrorsFromBulkFile(dh, transmission);

			    //go through error list, each one find the referenced sequence in the details list and set error message
			    Iterator<TransmitterErrorDetailGrpType> errorIt = errorList.iterator();

			    while (errorIt.hasNext()) {
			    	try {
				    	TransmitterErrorDetailGrpType curError = errorIt.next();
				    	String uniqueId = curError.getUniqueRecordId();

				    	//1095B-15-00000098|1|27
				    	if (uniqueId == null || uniqueId.isEmpty() || StringUtils.split(uniqueId, "|").length < 3) {
				    		logger.error("Invalid Unique Record Id encountered in IRS data file: " + uniqueId);
				    		continue;
				    	}

				    	int sequence = Integer.valueOf((StringUtils.split(uniqueId, "|")[2]));

				    	//if multiple errors, concat just the codes
				    	//one error, save code plus description
				    	List<ErrorMessageDetailType> errList = curError.getErrorMessageDetail();
				    	String recordErr = "";
				    	if (errList != null && errList.size() > 0) {
				    		if (errList.size() == 1) {
				    			ErrorMessageDetailType firstErr = errList.get(0);
				    			recordErr = firstErr.getErrorMessageCd() + ":" + firstErr.getErrorMessageTxt();
				    		} else {
				    			for (int i = 0; i < errList.size(); i++) {
				    				ErrorMessageDetailType cur = errList.get(i);
				    				recordErr += cur.getErrorMessageCd();
				    				recordErr += ":";
				    			}
				    		}

				    		if (recordErr.length() > 900) recordErr = recordErr.substring(0, 900) + " :TRUNCATED";

				    		//find the detail sequence and set the error
					    	Iterator<IrsTransmissionLogDetailEntry> set = transmission.getIrsTransmissionLogDetails().iterator();
					    	while (set.hasNext()) {
					    		IrsTransmissionLogDetailEntry cur = set.next();
					    		if (cur.getRecordSequence() == sequence) {
					    			cur.setErrorDescription(recordErr);
					    			break;
					    		}
					    	}
				    	}

			    	} catch (Exception ex) {
			    		logger.error("Error processing form data file records: " + ex.getMessage());
			    		continue;
			    	}
			    }
			} else {
				//status was returned but no match to any known type
				//should never happen
				transmission.setErrorDescription("Unknown Status Type Returned:" + respType.value());
				transmission.setTransmissionStatus((TransmissionStatus) this.lookupService.getByCode(
						TransmissionStatus.class, TransmissionStatus.FAILED.getCode()));
			}

		} catch (ServiceException | IOException e) {
			logger.error("RPC Exception in IRSAcknolegement request: " + e.getMessage());
		}

	 }

	/*
	 * Marshals the response data file attachment to
	 * xml response data and to an error detail list to process
	 *
	 */
	 @SuppressWarnings("unchecked")
	private List<TransmitterErrorDetailGrpType> getErrorsFromBulkFile(DataHandler dh, IrsTransmissionLogEntry transmission) {
		 List<TransmitterErrorDetailGrpType> errorList = null;
		 try {
		 	FormBCTrnsmtSubmissionDtlType errRoot =
					new FormBCTrnsmtSubmissionDtlType();

			us.gov.treasury.irs.msg.form1094_1095bctransmittermessage.ObjectFactory rFac =
					new us.gov.treasury.irs.msg.form1094_1095bctransmittermessage.ObjectFactory();

			JAXBContext context =
		        JAXBContext.newInstance(us.gov.treasury.irs.msg.form1094_1095bctransmittermessage.ObjectFactory.class);

			Unmarshaller m = context.createUnmarshaller();
		    Marshaller ms = context.createMarshaller();
		    ms.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
		    StringWriter sw = new StringWriter();

		    errRoot = ((JAXBElement<FormBCTrnsmtSubmissionDtlType>)
		    		m.unmarshal(dh.getInputStream())).getValue();

		    ms.marshal(rFac.createFormBCTransmitterSubmissionDtl(errRoot), sw);
		    String responseXml = sw.toString();
		    transmission.getIrsTransmissionLogData().iterator().next().setResponseData(responseXml);

		    errorList = errRoot.getACATransmitterSubmissionDetail().getTransmitterErrorDetailGrp();
		 } catch (JAXBException | IOException ex) {
			 logger.error("Error parsing bulk error data file from IRS: " + ex.getMessage());
		 }

		 return errorList;
	 }

	 /*
	  * Mock response for testing or internal validation only
	  * TODO move these mocks to external class and load through spring with configurable params
	  */
	 private ACABulkRequestTransmitterStatusDetailResponse mockResponse(String receipt) {
		 ACABulkRequestTransmitterStatusDetailResponse response =
				 new ACABulkRequestTransmitterStatusDetailResponse();

		 ACABulkReqTrnsmtStsRespGrpDtl det =
				 new ACABulkReqTrnsmtStsRespGrpDtl();

		 //TransmissionStatusCdType
		 us.gov.treasury.irs.ext.aca.air._7.ACABulkRequestTransmitterResponse type = new us.gov.treasury.irs.ext.aca.air._7.ACABulkRequestTransmitterResponse();

		 type.setTransmissionStatusCd(TransmissionStatusCodeType.ACCEPTED);
		 type.setErrorMessageDetail(null);
		 type.setReceiptId(receipt);

		 response.setACABulkRequestTransmitterResponse(type);


		 DataHandler handler = new DataHandler(new FileDataSource(this.getTempFileLocation() + "errorFile.xml"));
		 det.setBulkExchangeFile(handler);

		 response.setACABulkReqTrnsmtStsRespGrpDtl(det);

		 return response;
	 }


	/**
	 * Builds the xml payload for the entire form data file
	 * containing all 1094 and 1095B submissions in the batch request
	 * You should not invoke this service directly from outside context of the IRSTransmissionService
     * @param batch list of individual 1095B entries in the submission
     * @param correctionInd identifies in the batch contains Corrected or Replacement records for previous submissions
     * @param taxYear the tax year to process
     * @param originalReceipt the identifier of the original submissions that are being corrected or replaced
     */
	 public IrsTransmissionLogData buildRequestDataFile(List<IRS1095B> batch, String correctionInd, String taxYear, String originalReceipt) throws ServiceException {

		 Form1094BUpstreamDetailType transmitter = new Form1094BUpstreamDetailType();

		 us.gov.treasury.irs.msg.form1094_1095btransmitterupstreammessage.ObjectFactory f = new us.gov.treasury.irs.msg.form1094_1095btransmitterupstreammessage.ObjectFactory();

		 Form109495BTrnsmtUpstreamType rootContainer = f.createForm109495BTrnsmtUpstreamType();

		 //per IRS guide these values are required always to be the same
		 transmitter.setLineNum(BigInteger.valueOf(0));
		 transmitter.setRecordType("");

		 IrsTransmissionLogData requestData = new IrsTransmissionLogData();
		 String correctedInd = "0";
		 if (correctionInd.equalsIgnoreCase(SUBMIT_CORRECTION) || correctionInd.equalsIgnoreCase(SUBMIT_REPLACEMENT)) {
			 correctedInd = "1";
		 }


		 try {
			 //always only one 1094 in the bulk transmit
			 transmitter.setSubmissionId(PAYER_COUNT);

			 if (!correctionInd.equalsIgnoreCase(SUBMIT_ORIGINAL) && originalReceipt != null) {
				 transmitter.setOriginalUniqueSubmissionId(StringUtils.split(originalReceipt, "|")[0] + "|1");
			 }

			 transmitter.setTaxYr(DatatypeFactory.newInstance().newXMLGregorianCalendar(taxYear));
			 BusinessNameType busName = new BusinessNameType();
			 busName.setBusinessNameLine1Txt(LEGAL_NAME);
			 transmitter.setBusinessName(busName);
			 transmitter.setEmployerEIN(EIN);

			 OtherCompletePersonNameType name = new OtherCompletePersonNameType();
			 name.setPersonFirstNm(this.getIrsContactFname());
			 name.setPersonLastNm(this.getIrsContactLname());

			 transmitter.setContactNameGrp(name);
			 transmitter.setContactPhoneNum(this.getIrsContactPhone());


			 BusinessAddressGrpType busAddress = new BusinessAddressGrpType();
			 USAddressGrp us = new USAddressGrp();
			 us.setAddressLine1Txt(BUSINESS_STREET_ADDRESS);
			 us.setCityNm(BUSINESS_CITY);

			 us.setUSStateCd(StateType.fromValue(BUSINESS_STATE));
			 us.setUSZIPCd(BUSINESS_ZIP);
			 busAddress.setUSAddressGrp(us);

			 transmitter.setMailingAddressGrp(busAddress);

			 transmitter.setForm1095BAttachedCnt(BigInteger.valueOf(batch.size()));

			 //*******End 1094**************

			 //*********Start 1095 details list************
			 Iterator it = batch.iterator();
			 int pos = 1;
			 while (it.hasNext()) {
				 Form1095BUpstreamDetailType personDetail = new Form1095BUpstreamDetailType();
				 IRS1095B cur = (IRS1095B)it.next();

				 try {

					 personDetail.setLineNum(BigInteger.valueOf(0));
					 personDetail.setRecordType("");


					 personDetail.setRecordId((BigInteger.valueOf(pos)));
					 personDetail.setCorrectedInd(correctedInd);
					 personDetail.setTaxYr(DatatypeFactory.newInstance().newXMLGregorianCalendar(taxYear));

					 ResponsibleIndividualGrpType rig = new ResponsibleIndividualGrpType();

					 rig.setBirthDt(getDob(cur.getDOB()));

					 OtherCompletePersonNameType personName = new OtherCompletePersonNameType();
					 personName.setPersonFirstNm(cur.getFirstName());
					 personName.setPersonLastNm(cur.getLastName());
					 if (cur.getLastName() != null && !cur.getLastName().isEmpty()) {
						 personName.setPersonMiddleNm(cur.getMiddleName());
					 }
					 if (cur.getNameSuffix() != null && !cur.getNameSuffix().isEmpty()) {
						 personName.setSuffixNm(cur.getNameSuffix());
					 }

					 rig.setResponsibleIndividualName(personName);

					 rig.setTINRequestTypeCd(TINRequestTypeCodeType.INDIVIDUAL_TIN);
					 rig.setSSN(cur.getSSN());


					 if (!correctionInd.equalsIgnoreCase(SUBMIT_ORIGINAL) && originalReceipt != null) {
						 CorrectedRecordInfoGrpType correction = new CorrectedRecordInfoGrpType();
						 correction.setCorrectedUniqueRecordId(originalReceipt);
						 correction.setCorrectedRecordPayeeName(personName);
						 correction.setCorrectedRecordPayeeTIN(cur.getSSN());
						 personDetail.setCorrectedRecordInfoGrp(correction);
					 }

					 BusinessAddressGrpType personAddress = new BusinessAddressGrpType();

					 if (cur.getAddress().getCountry().equals(Country.CODE_USA.getName()) ||
							 (cur.getAddress().getState() != null && cur.getAddress().getZipCode() != null)) {
						 USAddressGrp add = new USAddressGrp();
						 add.setAddressLine1Txt(cur.getAddress().getLine1());
						 if (cur.getAddress().getLine2() != null && !cur.getAddress().getLine2().isEmpty()) {
							 add.setAddressLine2Txt(cur.getAddress().getLine2());
						 }
						 add.setCityNm(cur.getAddress().getCity());
						 add.setUSStateCd(StateType.fromValue(cur.getAddress().getState()));
						 add.setUSZIPCd(cur.getAddress().getZipCode());
						 if (cur.getAddress().getZipPlus4() != null && !cur.getAddress().getZipPlus4().isEmpty()) {
							 add.setUSZIPExtensionCd(cur.getAddress().getZipPlus4());
						 }

						 personAddress.setUSAddressGrp(add);
					 } else {
						 ForeignAddressGrp add = new ForeignAddressGrp();
						 add.setAddressLine1Txt(cur.getAddress().getLine1());
						 if (cur.getAddress().getLine2() != null && !cur.getAddress().getLine2().isEmpty()) {
							 add.setAddressLine2Txt(cur.getAddress().getLine2());
						 }
						 add.setCityNm(cur.getAddress().getCity());
						 add.setCountryNm(cur.getAddress().getCountry());
						 add.setForeignPostalCd(cur.getAddress().getPostalCode());
						 if (cur.getAddress().getProvince() != null && !cur.getAddress().getProvince().isEmpty()) {
							 add.setForeignProvinceNm(cur.getAddress().getProvince());
						 }

						 personAddress.setForeignAddressGrp(add);
					 }


					 rig.setMailingAddressGrp(personAddress);

					 rig.setPolicyOriginCd("C");

					 personDetail.setResponsibleIndividualGrp(rig);

					 //Issuer Info Group
					 IssuerInfoGrpType issuer = new IssuerInfoGrpType();
					 issuer.setBusinessName(busName);
					 issuer.setContactPhoneNum(BUSINESS_PHONE);
					 issuer.setEIN(EIN);
					 issuer.setMailingAddressGrp(busAddress);
					 issuer.setTINRequestTypeCd(TINRequestTypeCodeType.BUSINESS_TIN);

					 personDetail.setIssuerInfoGrp(issuer);

					 //Covered Individual
					 EmployerCoveredIndividualType personInfo = new EmployerCoveredIndividualType();

					 personInfo.setBirthDt(getDob(cur.getDOB()));
					 personInfo.setCoveredIndividualName(personName);
					 personInfo.setPersonNameControlTxt(cur.getLastName());
					 personInfo.setSSN(cur.getSSN());
					 personInfo.setTINRequestTypeCd(TINRequestTypeCodeType.INDIVIDUAL_TIN);

					 if (cur.getMecPeriod().isCoveredAll12Months()) {
						 personInfo.setCoveredIndividualAnnualInd("1");
					 } else {
						 personInfo.setCoveredIndividualAnnualInd("0");
					 }

					 MonthIndGrpType period = new MonthIndGrpType();

					 period.setJanuaryInd((boolean) cur.getMecPeriod().getCoverageMonths().containsKey("1") && (boolean) cur.getMecPeriod().getCoverageMonths().get("1") ? "1" : "0");
					 period.setFebruaryInd((boolean) cur.getMecPeriod().getCoverageMonths().containsKey("2") && (boolean) cur.getMecPeriod().getCoverageMonths().get("2") ? "1" : "0");
					 period.setMarchInd((boolean) cur.getMecPeriod().getCoverageMonths().containsKey("3") && (boolean) cur.getMecPeriod().getCoverageMonths().get("3") ? "1" : "0");
					 period.setAprilInd((boolean) cur.getMecPeriod().getCoverageMonths().containsKey("4") && (boolean) cur.getMecPeriod().getCoverageMonths().get("4") ? "1" : "0");
					 period.setMayInd((boolean) cur.getMecPeriod().getCoverageMonths().containsKey("5") && (boolean) cur.getMecPeriod().getCoverageMonths().get("5") ? "1" : "0");
					 period.setJuneInd((boolean) cur.getMecPeriod().getCoverageMonths().containsKey("6") && (boolean) cur.getMecPeriod().getCoverageMonths().get("6") ? "1" : "0");
					 period.setJulyInd((boolean) cur.getMecPeriod().getCoverageMonths().containsKey("7") && (boolean) cur.getMecPeriod().getCoverageMonths().get("7") ? "1" : "0");
					 period.setAugustInd((boolean) cur.getMecPeriod().getCoverageMonths().containsKey("8") && (boolean) cur.getMecPeriod().getCoverageMonths().get("8") ? "1" : "0");
					 period.setSeptemberInd((boolean) cur.getMecPeriod().getCoverageMonths().containsKey("9") && (boolean) cur.getMecPeriod().getCoverageMonths().get("9") ? "1" : "0");
					 period.setOctoberInd((boolean) cur.getMecPeriod().getCoverageMonths().containsKey("10") && (boolean) cur.getMecPeriod().getCoverageMonths().get("10") ? "1" : "0");
					 period.setNovemberInd((boolean) cur.getMecPeriod().getCoverageMonths().containsKey("11") && (boolean) cur.getMecPeriod().getCoverageMonths().get("11") ? "1" : "0");
					 period.setDecemberInd((boolean) cur.getMecPeriod().getCoverageMonths().containsKey("12") && (boolean) cur.getMecPeriod().getCoverageMonths().get("12") ? "1" : "0");

					 personInfo.setCoveredIndividualMonthlyIndGrp(period);

					 personDetail.getCoveredIndividualGrp().add(personInfo);


				 } catch (Exception ex) {
					 //something goes wrong before it's added, remove it from the batch
					 StringWriter sw = new StringWriter();
					 ex.printStackTrace(new PrintWriter(sw));
					 String exceptionDetails = sw.toString();
					 logger.error("Unknown Exception in building IRS data file entry for entry: " + cur + " " + exceptionDetails);
					 it.remove();
					 continue;
				 }

				 transmitter.getForm1095BUpstreamDetail().add(personDetail);
				 pos++;
			 }

			 rootContainer.getForm1094BUpstreamDetail().add(transmitter);

			 String xml = buildXmlDocument(f.createForm109495BTransmittalUpstream(rootContainer));

			 requestData.setRequestData(xml);


		} catch (DatatypeConfigurationException e) {
			logger.error("DataTypeError in BuildDataFile:" + e.getMessage(), e);
		} catch (Exception ex) {
			logger.error("Error in BuildDataFile:" + ex.getMessage(), ex);
		}

		return requestData;

	 }
	 /*
	  * Utility for converting imprecise dob to xml
	  *
	  */
	 private XMLGregorianCalendar getDob(String dob) throws ParseException, DatatypeConfigurationException {
		Date dt = null;
		if (dob != null && dob.length() > 3) {
			 if (dob.length() == 8) {
				 SimpleDateFormat f = new SimpleDateFormat("yyyyMMdd");
				 dt = f.parse(dob);
			 } else if (dob.length() == 6) {
				 SimpleDateFormat f = new SimpleDateFormat("yyyyMM");
				 dt = f.parse(dob);
			 } else {
				 SimpleDateFormat f = new SimpleDateFormat("yyyy");
				 dt = f.parse(dob);
			 }
			 GregorianCalendar cal = new GregorianCalendar();
			 cal.setTime(dt);


			 XMLGregorianCalendar xmlDate = DatatypeFactory.newInstance().newXMLGregorianCalendarDate
					 (cal.get(Calendar.YEAR), cal.get(Calendar.MONTH)+1, cal.get(Calendar.DAY_OF_MONTH), 0);
			 xmlDate.setTimezone(DatatypeConstants.FIELD_UNDEFINED);

			 return xmlDate;
		 }
		 return null;
	 }

 	/**
	 * Finds single person entry from stored bulk transmission payload entry
	 * and unmarshals xml to usable object rest of the system can use
     * @param dataFile the dataFile entry containing the original CLOB xml data for the transmission
     * @param sequence the record sequence in this data file entry that is to be retrieved
     */
	 @SuppressWarnings("unchecked")
	public IRS1095B convertSingleEntry(IrsTransmissionLogData dataFile, int sequence) throws ServiceException {

	 	String xml = dataFile.getRequestData();

		Form1095BUpstreamDetailType entry = null;
		IRS1095B submission = null;
		MECPeriod period = new MECPeriod();
		HashMap coverageMonths = new HashMap();

		try {

			if (xml != null && !xml.isEmpty()) {

				Form109495BTrnsmtUpstreamType result = JAXB.unmarshal(new StringReader(xml), Form109495BTrnsmtUpstreamType.class);

				//always only one 1094 entry
				List<Form1095BUpstreamDetailType> details = result.getForm1094BUpstreamDetail().get(0).getForm1095BUpstreamDetail();

				for (Form1095BUpstreamDetailType object : details) {
			        if (object.getRecordId().intValue() == sequence) {
			            entry = object;
			            break;
			        }
			    }

				if (entry != null) {

					submission = new IRS1095B();

					SimpleAddress address = new SimpleAddress();
					if (entry.getResponsibleIndividualGrp().getMailingAddressGrp() != null) {

						USAddressGrp add = entry.getResponsibleIndividualGrp().getMailingAddressGrp().getUSAddressGrp();

						if (add != null) {

							address.setCity(add.getCityNm());
							address.setLine1(add.getAddressLine1Txt());
							address.setLine2(add.getAddressLine2Txt());
							address.setState(add.getUSStateCd().name());
							address.setZipCode(add.getUSZIPCd());
							address.setZipPlus4(add.getUSZIPExtensionCd());
						} else {
							ForeignAddressGrp fadd = entry.getResponsibleIndividualGrp().getMailingAddressGrp().getForeignAddressGrp();
							if (fadd != null) {
								address.setCity(fadd.getCityNm());
								address.setCountry(fadd.getCountryNm());
								address.setLine1(fadd.getAddressLine1Txt());
								address.setLine2(fadd.getAddressLine2Txt());
								address.setPostalCode(fadd.getForeignPostalCd());
								address.setProvince(fadd.getForeignProvinceNm());
							}
						}


						submission.setAddress(address);
					}

					period.setCoveredAll12Months(entry.getCoveredIndividualGrp().get(0).getCoveredIndividualAnnualInd().equalsIgnoreCase("0") ? false : true);
					coverageMonths.put("1", (entry.getCoveredIndividualGrp().get(0).getCoveredIndividualMonthlyIndGrp().getJanuaryInd().equalsIgnoreCase("1") ? true : false));
					coverageMonths.put("2", (entry.getCoveredIndividualGrp().get(0).getCoveredIndividualMonthlyIndGrp().getFebruaryInd().equalsIgnoreCase("1") ? true : false));
					coverageMonths.put("3", (entry.getCoveredIndividualGrp().get(0).getCoveredIndividualMonthlyIndGrp().getMarchInd().equalsIgnoreCase("1") ? true : false));
					coverageMonths.put("4", (entry.getCoveredIndividualGrp().get(0).getCoveredIndividualMonthlyIndGrp().getAprilInd().equalsIgnoreCase("1") ? true : false));
					coverageMonths.put("5", (entry.getCoveredIndividualGrp().get(0).getCoveredIndividualMonthlyIndGrp().getMayInd().equalsIgnoreCase("1") ? true : false));
					coverageMonths.put("6", (entry.getCoveredIndividualGrp().get(0).getCoveredIndividualMonthlyIndGrp().getJuneInd().equalsIgnoreCase("1") ? true : false));
					coverageMonths.put("7", (entry.getCoveredIndividualGrp().get(0).getCoveredIndividualMonthlyIndGrp().getJulyInd().equalsIgnoreCase("1") ? true : false));
					coverageMonths.put("8", (entry.getCoveredIndividualGrp().get(0).getCoveredIndividualMonthlyIndGrp().getAugustInd().equalsIgnoreCase("1") ? true : false));
					coverageMonths.put("9", (entry.getCoveredIndividualGrp().get(0).getCoveredIndividualMonthlyIndGrp().getSeptemberInd().equalsIgnoreCase("1") ? true : false));
					coverageMonths.put("10", (entry.getCoveredIndividualGrp().get(0).getCoveredIndividualMonthlyIndGrp().getOctoberInd().equalsIgnoreCase("1") ? true : false));
					coverageMonths.put("11", (entry.getCoveredIndividualGrp().get(0).getCoveredIndividualMonthlyIndGrp().getNovemberInd().equalsIgnoreCase("1") ? true : false));
					coverageMonths.put("12", (entry.getCoveredIndividualGrp().get(0).getCoveredIndividualMonthlyIndGrp().getDecemberInd().equalsIgnoreCase("1") ? true : false));

					period.setCoverageMonths(coverageMonths);
					submission.setMecPeriod(period);

				}

			}

		} catch (Exception e) {
			logger.error("Error converting XML data request to object: " + e.getMessage());
		}

		 return submission;
	 }

	public LookupService getLookupService() {
		return lookupService;
	}

	public void setLookupService(LookupService lookupService) {
		this.lookupService = lookupService;
	}

	public String getIrsProcessingCode() {
		return irsProcessingCode;
	}

	public void setIrsProcessingCode(String irsProcessingCode) {
		this.irsProcessingCode = irsProcessingCode;
	}

	public DAOOperations getGenericDAO() {
		return genericDAO;
	}

	public void setGenericDAO(DAOOperations genericDAO) {
		this.genericDAO = genericDAO;
	}

	public String getIrsSignatureFile() {
		return irsSignatureFile;
	}

	public void setIrsSignatureFile(String irsSignatureFile) {
		this.irsSignatureFile = irsSignatureFile;
	}

	public String getTransmissionControlCode() {
		return transmissionControlCode;
	}

	public void setTransmissionControlCode(String transmissionControlCode) {
		this.transmissionControlCode = transmissionControlCode;
	}

	public String getSoftwareId() {
		return softwareId;
	}

	public String getIrsContactFname() {
		return irsContactFname;
	}

	public void setIrsContactFname(String irsContactFname) {
		this.irsContactFname = irsContactFname;
	}

	public String getIrsContactLname() {
		return irsContactLname;
	}

	public void setIrsContactLname(String irsContactLname) {
		this.irsContactLname = irsContactLname;
	}

	public String getIrsContactPhone() {
		return irsContactPhone;
	}

	public void setIrsContactPhone(String irsContactPhone) {
		this.irsContactPhone = irsContactPhone;
	}

	public void setSoftwareId(String systemId) {
		this.softwareId = systemId;
	}

	/*
	 * sets all common manifest header elements
	 *
	 */
	 private void setCommonHeader(IrsTransmissionLogEntry transmission, ACATransmitterManifestReqDtl headerManifest) {

		 try {

			 headerManifest.setPaymentYr(DatatypeFactory.newInstance().newXMLGregorianCalendar(transmission.getTaxYear()));

			 //set 0 as current filing year, 1 as correction for previous year
			 Calendar prevYear = Calendar.getInstance();
			 prevYear.add(Calendar.YEAR, -1);
			 int curTaxYear = prevYear.get(Calendar.YEAR);
			 if (Integer.valueOf(transmission.getTaxYear()) < curTaxYear) {
				 headerManifest.setPriorYearDataInd("1");
			 } else {
				 headerManifest.setPriorYearDataInd("0");
			 }

			 //EIN
			 headerManifest.setEIN(EIN);

			 //O, C, R
			 if (transmission.getCorrectionInd().equalsIgnoreCase(SUBMIT_ORIGINAL)) {
				 headerManifest.setTransmissionTypeCd(TransmissionTypeCdType.O);
			 } else if (transmission.getCorrectionInd().equalsIgnoreCase(SUBMIT_CORRECTION)) {
				 headerManifest.setTransmissionTypeCd(TransmissionTypeCdType.C);
			 } else {
				 headerManifest.setTransmissionTypeCd(TransmissionTypeCdType.R);
			 }

			 //Test or Production Indicator
			 headerManifest.setTestFileCd(this.getIrsProcessingCode());

			 //Transmitter Name
			 BusinessNameType busName = new BusinessNameType();
			 busName.setBusinessNameLine1Txt(LEGAL_NAME);
			 headerManifest.setTransmitterNameGrp(busName);

			 OtherCompletePersonNameType name = new OtherCompletePersonNameType();
			 name.setPersonFirstNm(this.getIrsContactFname());
			 name.setPersonLastNm(this.getIrsContactLname());

			 BusinessAddressGrpType busAddress = new BusinessAddressGrpType();
			 USAddressGrp us = new USAddressGrp();
			 us.setAddressLine1Txt(BUSINESS_STREET_ADDRESS);
			 us.setCityNm(BUSINESS_CITY);
			 us.setUSStateCd(StateType.fromValue(BUSINESS_STATE));

			 us.setUSZIPCd(BUSINESS_ZIP);
			 busAddress.setUSAddressGrp(us);

			 CompanyInformationGrp company = new CompanyInformationGrp();
			 company.setCompanyNm(LEGAL_NAME);
			 company.setContactNameGrp(name);
			 company.setContactPhoneNum(this.getIrsContactPhone());
			 company.setMailingAddressGrp(busAddress);
			 headerManifest.setCompanyInformationGrp(company);

			 VendorInformationGrp inhouse = new VendorInformationGrp();
			 inhouse.setContactNameGrp(name);
			 inhouse.setContactPhoneNum(this.getIrsContactPhone());
			 inhouse.setVendorCd("I");
			 headerManifest.setVendorInformationGrp(inhouse);

			 //counts
			 headerManifest.setTotalPayerRecordCnt(PAYER_COUNT);

			 headerManifest.setTotalPayeeRecordCnt(BigInteger.valueOf(transmission.getIrsTransmissionLogDetails().size()));

			 //sw ID
			 headerManifest.setSoftwareId(this.getSoftwareId());

			 headerManifest.setFormTypeCd(FORM_TYPE_CODE);

			 //headerManifest.setBinaryFormatCd(BinaryFormatCodeType.APPLICATION_XML);
			 headerManifest.setBinaryFormatCd(TYPE_XML);


		} catch (Exception e) {
			logger.error("Data Type Error in IRS Transmit: ", e);
		}

	 }

	public String getClientSystemId() {
		return clientSystemId;
	}

	public void setClientSystemId(String clientSystemId) {
		this.clientSystemId = clientSystemId;
	}

	public String getTempFileLocation() {
		return tempFileLocation;
	}

	public void setTempFileLocation(String tempFileLocation) {
		this.tempFileLocation = tempFileLocation;
	}

	public String getIrsServiceBulkEndPoint() {
		return irsServiceBulkEndPoint;
	}

	public void setIrsServiceBulkEndPoint(String irsServiceBulkEndPoint) {
		this.irsServiceBulkEndPoint = irsServiceBulkEndPoint;
	}

	public String getIrsServiceStatusEndPoint() {
		return irsServiceStatusEndPoint;
	}

	public void setIrsServiceStatusEndPoint(String irsServiceStatusEndPoint) {
		this.irsServiceStatusEndPoint = irsServiceStatusEndPoint;
	}

}
